# IMPORTS
from datetime import datetime, timedelta
import numpy as np
import pandas as pd
from pulp import *

# SUPPORT FUNCTIONS
# Get index ranges for different months in data
def getMonthIndexes(months):
    indexes = []
    
    indexes.append(0)
    
    currentMonth = months[0]
    for i in range(len(months)):
        if months[i] != currentMonth:
            currentMonth = months[i]
            indexes.append(i)
    
    indexes.append(len(months))
    
    return indexes

# Get index ranges for different days in data
def getDayIndexes(days):
    indexes = []
    
    indexes.append(0)
    
    currentDay = days[0]
    for i in range(len(days)):
        if days[i] != currentDay:
            currentDay = days[i]
            indexes.append(i)
    
    indexes.append(len(days))
    
    return indexes

# LOAD DATA
hourly_data_path = "importfile/StOlavsveiResultDf.csv"

daily_data_path = "importfile/ControllableLoadsStOlavs_withoutCar_Heater.csv" #case D: only use for no load control


#variables 
batteryPower = 3.3
batteryEnergy = 10
carPower = 0
heatingPower = 0
installedPv = 1/1000 #because the PV production is in W, not kW
importLimit = 20
exportLimit = 20



hourly_df = pd.read_csv(hourly_data_path, delimiter=",")
daily_df = pd.read_csv(daily_data_path, delimiter=",")

# MILP


def runMILP(size):
    S_max = size[0]
    Q_max = size[1]
    e = [] # export
    i = [] # import
    s = [] # SOC
    p = [] # monthly peaks
    b_p = {} # 2d dict of dv for which effektledd "trinn" it falls under

    wh = [] # dv for water heater
    ev = [] # dv for electrical vehicle



    S_start = 0

    T_i = hourly_df['Tot price in'].copy().values # spot + nettleie etc for import
    T_e = hourly_df['Spot'].copy().values # spot + innmatingstariff for export
    
    # P = (hourly_df['Produced west']*installedPv).copy().values # PV production
    # P = (hourly_df['Produced east']*installedPv).copy().values # PV production
    # P = (hourly_df['Produced south']*installedPv).copy().values # PV production
    P = ((hourly_df['Produced west'] + hourly_df['Produced east'])*installedPv).copy().values # PV production
    
    L = (hourly_df['Electric car [kWh]'] + hourly_df['Heating tank [kWh]'] + hourly_df['Other loads [kWh]']).copy().values #case D
  

    L_wh = daily_df['Heating tank [kWh]'].copy().values # daily load for water heater
    L_ev = daily_df['Electric car [kWh]'].copy().values # daily load for EV
    D = len(L_wh) # number of days in data
    H_d = getDayIndexes(hourly_df['Day of Year'].copy().values)

    H_m = getMonthIndexes(hourly_df['Month'].copy().values)
    M = len(H_m)-1


    EF_T = [0, 2, 5, 10, 15, 20, 25]
    EF = [125, 206, 350, 494, 638, 781]
    N_T = len(EF) # number of "trinn"
    B = 0.00001 # very small value
        
    N = len(T_i)

    # MILP FORMULATION
    print("MILP")

    problem = LpProblem('MILP', LpMinimize)

    for ind in range(N):
        i.append(LpVariable('i_'+str(ind), lowBound=0,upBound=None, cat='Continuous'))
        s.append(LpVariable('s_'+str(ind), lowBound=0, upBound=S_max, cat='Continuous'))  
        wh.append(LpVariable('wh'+str(ind), lowBound=0, upBound=heatingPower, cat='Continuous')) # max that can be used for heating power
        ev.append(LpVariable('ev'+str(ind), lowBound=0, upBound=carPower, cat='Continuous')) # max that can be used for car charging

        if (P[ind] - L[ind] + Q_max) > 0:
            e.append(LpVariable('e_'+str(ind), lowBound=0, upBound=(P[ind] - L[ind] + Q_max), cat='Continuous'))
        else:  
            e.append(LpVariable('e_'+str(ind), lowBound=0, upBound=0, cat='Continuous'))


    for m in range(M):
        p.append(LpVariable('p'+str(m), lowBound=0,upBound=None,cat='Continuous'))
        for t in range(N_T):
            b_p[(m,t)] = LpVariable('b_p'+str(m)+str(t), cat='Binary')


    # Objective Function (Must be first thing added to problem?)
    problem += lpSum([(T_i[ind] + B)*i[ind] - (T_e[ind])*e[ind] for ind in range(N)]) \
                + lpSum([b_p[(m,t)]*EF[t] for m in range(M) for t in range(N_T)]), 'Objective Function'

    # Constraints
    problem += S_start == s[N-1], 'End SOC Constraint'


    problem += s[0] - S_start >= -Q_max, 'S_i - S_i-1 Lower Constraint0'
    problem += s[0] - S_start <= Q_max, 'S_i - S_i-1 Upper Constraint0'
    problem += -e[0] + i[0] == L[0] - P[0] + s[0] - S_start + wh[0] + ev[0], 'Balancing Constraint0'
    problem += i[0] <= importLimit, 'Limit import0'
    problem += e[0] <= exportLimit, 'Limit export0'
    for ind in range(1,N,1):
        problem += s[ind] - s[ind-1] >= -Q_max, 'S_i - S_i-1 Lower Constraint' +str(ind)
        problem += s[ind] - s[ind-1] <= Q_max, 'S_i - S_i-1 Upper Constraint' +str(ind)

        problem += -e[ind] + i[ind] == L[ind] - P[ind] + s[ind] - s[ind-1] + wh[ind] + ev[ind], 'Balancing Constraint' + str(ind)

        problem += i[ind] <= importLimit, 'Limit import'+str(ind)
        problem += e[ind] <= exportLimit, 'Limit export'+str(ind)


    for m in range(M):
        problem += lpSum([b_p[(m,t)] for t in range(N_T)]) == 1, 'Only select one trinn constraint'+str(m)
        problem += p[m] >= lpSum([EF_T[t]*b_p[(m,t)] for t in range(N_T-1)]), "above trinn"+str(m)
        problem += p[m] <= lpSum([EF_T[t+1]*b_p[(m,t)] for t in range(N_T-1)]), "below trinn"+str(m)
        for ind in range(H_m[m], H_m[m+1], 1):
            problem += p[m] >= i[ind], 'Peak Constraint'+str(m)+str(ind)

    for d in range(D):
        problem += lpSum([wh[ind] for ind in range(H_d[d], H_d[d+1], 1)]) == L_wh[d], 'Water heater daily constrint'+str(d)
        problem += lpSum([ev[ind] for ind in range(H_d[d], H_d[d+1], 1)]) == L_ev[d], 'Electrical vehicle daily constrint'+str(d)

    # solve problem and print results
    status = problem.solve(PULP_CBC_CMD(msg=0))
    print(LpStatus[status])
    #for v in problem.variables():
        #print(v.name, "=", v.varValue)

    #print("q_0 = " + str(s[0].varValue - S_start))
    #for i in range(1,N,1):
        #print("q_"+str(i)+ " = " + str(s[i].varValue - s[i-1].varValue))

    # peaks
    for ind in range(12):
        print("Month: " + str(ind+1) + " Peak: " + str(p[ind].varValue))

        for t in range(N_T):
            if b_p[(ind,t)].varValue:
                print('Additional cost: ' + str(EF[t]))

    print("Objective Function")
    #print(problem.objective)

    print("Objective Function Value")
    print(value(problem.objective))
    
    print("import")
    sum_i = 0
    sum_e = 0
    
    for ind in range(N):
        sum_i += i[ind].varValue
        sum_e += e[ind].varValue
        
    print(sum_i)
    print("export")
    print(sum_e)
    # print("sum")
    # print(sum + sum_e)
    
    sum_cycle = 0
    sum_cycle += abs(s[0].varValue-S_start)
    for m in range(1,N):
        sum_cycle += abs(s[m].varValue-s[m-1].varValue)
    sum_cycle = sum_cycle/2/max(S_max,1)
    
    print("cycles")
    print(sum_cycle)
    
    # for m in range(M):
    #     spot_e = 0
    #     spot_i = 0
    #     for ind in range(H_m[m], H_m[m+1], 1):
    #         spot_i += i[ind].varValue * T[ind]
    #         spot_e += e[ind].varValue * T[ind]
    #     print("Month: " + str(m+1) + " spot: "+str(spot_i))
    #     print("Month: " + str(m+1) + " spot export: " + str(spot_e))
    
    # print(sum([im for im in i]))
    
    return value(problem.objective)

runMILP((batteryEnergy, batteryPower))
#runMILP((0, 0))
